//------------------------------------------------------------------------------
// Filename: dlog64sceil.c
//   Author: Adriano Arce
//     Date: September 3, 2015
//------------------------------------------------------------------------------
// Direct Dependencies:
//     polyutil.o: For Z_p utilities.
//------------------------------------------------------------------------------
// To compile this into an object file, use:
//     cc -Wall -O3 -c dlog64sceil.c
//------------------------------------------------------------------------------
#include <string.h>

#define LONG long long int
#define FAIL -1

#define NUM_UNSIGNED_BITS 63

#define MAX_NUM_PRIMES NUM_UNSIGNED_BITS
#define SHANKS_THRESHOLD 10

#define NUM_INT_VALS 1
#define NUM_LONG_VALS 2
#define NUM_LONG_ARRS 7
#define NUM_INT_PTR_ARRS 1
#define NUM_LONG_PTR_ARRS 1

//------------------------------------------------------------------------------
// Z_p utilities (from polyutil).
//------------------------------------------------------------------------------

#define add32s logadd32s
#define sub32s logsub32s
#define mul64s logmul64s
inline int add32s(int a, int b, int p)
{
    int t = (a - p) + b;
    t += (t >> 31) & p;
    return t;
}
inline int sub32s(int a, int b, int p)
{
    int t = a - b;
    t += (t >> 31) & p;
    return t;
}
inline LONG mul64s(LONG a, LONG b, LONG p)
{
    LONG q, r;
    __asm__ __volatile__(           \
    "       mulq    %%rdx           \n\t" \
    "       divq    %4              \n\t" \
    : "=a"(q), "=d"(r) : "0"(a), "1"(b), "rm"(p));
    return r;
}

LONG modinv64s(LONG c, LONG p);
LONG powmod64s(LONG a, LONG n, LONG p);
int ceilsqrt32s(int n);

//------------------------------------------------------------------------------
// Other Z_p utilities. TODO: Move these implementations to polyutil.
//------------------------------------------------------------------------------

/*
 * Precomputes M and I for chrem64s().
 *
 * INPUT:
 *     Q: The pairwise relatively prime moduli whose product is n.
 *     n: The product of the moduli (the order of alpha).
 *     M: The reduced moduli n / Q[i].
 *     I: The inverses 1 / M[i] mod Q[i].
 *     k: The size of Q, M, and I.
 */
void setupChrem64s(LONG* Q, LONG n, LONG* M, LONG* I, int k)
{
    int i;
    LONG q, m;
    for (i = 0; i < k; i++)
    {
        q = Q[i];
        m = n / q;
        I[i] = modinv64s(m, q);
        M[i] = m;
    }
}

/*
 * Applies the Chinese Remainder Theorem to solve u = U[i] (mod Q[i]).
 *
 * PRECONDITION:
 *     Assumes that setupChrem64s() has already been called to set up M and I.
 *
 * INPUT:
 *     U: The evaluation points U[i] = u mod Q[i].
 *     Q: The pairwise relatively prime moduli whose product is n.
 *     n: The product of the moduli (the order of alpha).
 *     M: The reduced moduli n / Q[i].
 *     I: The inverses 1 / M[i] mod Q[i].
 *     k: The size of U, Q, M, and I.
 */
LONG chrem64s(LONG* U, LONG* Q, LONG n, LONG* M, LONG* I, int k)
{
    int i;
    LONG u, q, t;
    
    u = 0;
    for (i = 0; i < k; i++)
    {
        q = Q[i];
        t = mul64s(U[i], I[i], q);
        u -= t * M[i];
        u += (u >> NUM_UNSIGNED_BITS) & n;
    }
    
    if (u != 0)
        u = n - u;
    
    return u;  // 0 <= u < n
}

/*
 * Naively solves alpha^x = y mod p for 0 <= x < n.
 */
int logNaive64s(LONG y, LONG alpha, LONG p, int n)
{
    int x;
    LONG a;
    a = 1;
    for (x = 0; x < n; x++)
    {
        if (a == y)
            return x;
        a = mul64s(a, alpha, p);
    }
    return FAIL;
}

//------------------------------------------------------------------------------
// The main discrete logarithm implementation.
//------------------------------------------------------------------------------

/*
 * Prepares the hash tables for Shanks' algorithm.
 *
 * INPUT:
 *     alpha: The base of the logarithm.
 *         p: The size of the finite field Z_p.
 *         n: The order of alpha in Z_p.
 *         Y: The hash table for [1, alpha^s, alpha^{2s}, ..., alpha^{(s-1)s}].
 *         X: The hash table for [0,       s,         2s, ...,         (s-1)s].
 *
 * NOTE:
 *     Y stores the keys that get hashed (modulo t).
 *     X stores the values in each key-value pair (alpha^{js}, js).
 *     t = 2s + 1 means that each of Y and X will be half full.
 */
void hashShanks64s(LONG alpha, LONG p, int n, LONG* Y, int* X)
{
    int s, t, i, j, x, h;
    LONG y, z;
    s = ceilsqrt32s(n);
    t = 2*s + 1;
    
    // Compute y = alpha^s mod p, taking baby steps of size alpha.
    y = alpha;
    for (i = 1; i < s; i++)
        y = mul64s(alpha, y, p);
    
    // Since 0 is not in <alpha>, we can safely initialize Y with 0's.
    memset(Y, 0, t * sizeof(Y));

    // The first hash is h = z % t = 1 % t = 1.
    Y[1] = 1;  // z = alpha^x = alpha^0 = 1.
    X[1] = 0;

    z = y;
    for (j = 1, x = s; j < s; x += s, j++)  // x = js for j from 1 to s - 1.
    {
        // Perform linear probing until we find the first nonzero element in Y.
        for (h = z % t; Y[h]; h = add32s(h, 1, t));
        
        Y[h] = z;  // z = y^x = alpha^{js}.
        X[h] = x;
        
        z = mul64s(y, z, p);  // Take a giant step of size y = alpha^s.
    }
}

/*
 * Computes x = log[alpha](y) in Z_p via Shanks' algorithm.
 *
 * PRECONDITION:
 *     Assumes that hashShanks64s() has already been called to set up Y and X.
 *
 * INPUT:
 *         y: The argument of the logarithm (y = alpha^x).
 *     alpha: The base of the logarithm.
 *         p: The size of the finite field Z_p.
 *         n: The order of alpha in Z_p.
 *         Y: The hash table for [1, alpha^s, alpha^{2s}, ..., alpha^{(s-1)s}].
 *         X: The hash table for [0,       s,         2s, ...,         (s-1)s].
 *
 * NOTE:
 *     Consider <alpha> as a subgroup of Z*_p = {1, ..., p-1}.
 *     Then log[alpha]: <alpha> -> Z_n = {0, ..., n-1} is a group isomorphism.
 *     So the discrete log, if it exists, is unique modulo n.
 */
int logShanks64s(LONG y, LONG alpha, LONG p, int n, LONG* Y, int* X)
{
    int i, s, t, h;
    s = ceilsqrt32s(n);
    t = 2*s + 1;
    for (i = 0; i < s; i++)
    {
        // Perform linear probing until we find the first nonzero element in Y.
        // If alpha^{js} = y*alpha^i, then alpha^x = y = alpha^{js - i}.
        // Hence, since alpha^n = 1, we return x = js - i (mod n).
        for (h = y % t; Y[h]; h = add32s(h, 1, t))
            if (Y[h] == y)
                return sub32s(X[h], i, n);
        
        y = mul64s(alpha, y, p);  // Take a baby step of size alpha.
    }
    return FAIL;  // Such a log does not exist.
}

/*
 * Computes x = log[beta](y) mod q^c in Z_p via the Pohlig-Hellman algorithm.
 *
 * INPUT:
 *         y: The argument of the logarithm (y = beta^x).
 *     delta: A power of the base (delta = beta^{q^{c - 1}} = alpha^{n/q}).
 *         p: The size of the finite field Z_p.
 *     gamma: The inverse of beta mod p (gamma = beta^{-1} = alpha^{-m}).
 *         n: The order of alpha.
 *         q: A prime factor of n = |alpha|.
 *         c: The exponent of q in the prime factorization of n.
 *         Y: The hash table for [1, delta^s, delta^{2s}, ..., delta^{(s-1)s}].
 *         X: The hash table for [0,       s,         2s, ...,         (s-1)s].
 */
LONG pohligHellman64s(LONG y, LONG delta, LONG p, LONG gamma, LONG n, int q,
                      int c, LONG* Y, int* X)
{
    int j, xj;
    LONG z, x, qj, k, w;
    z  = y;
    x  = 0;
    qj = 1;
    k  = n/q;
    for (j = 0; j < c; j++)
    {
        //----------------------------------------------------------------------
        // Loop invariants at the beginning of iteration j.
        //----------------------------------------------------------------------
        //     gamma := beta^{-q^j};
        //         z := beta^{x_j*q^j + ... + x_{c - 1}*q^{c - 1}};
        //         x := x_0 + x_1*q + x_2*q^2 + ... + x_{j - 1}*q^{j - 1};
        //        qj := q^j;
        //         k := n/q^{j + 1} = q^{c - j - 1};
        //----------------------------------------------------------------------
        
        // Solve delta^{x_j} = z^{q^{c - j - 1}} for 0 <= x_j < q.
        // Note that since delta = alpha^{n/q}, we know that |delta| = q.
        w = powmod64s(z, k, p);
        if (q == 2)  // Optimize for q = 2.
        {
            if (w == 1)
                xj = 0;
            else if (w == delta)
                xj = 1;
            else
                xj = FAIL;
        }
        else if (q < SHANKS_THRESHOLD)
            xj = logNaive64s(w, delta, p, q);
        else
            xj = logShanks64s(w, delta, p, q, Y, X);
        
        if (xj == FAIL)
            return FAIL;  // Such a log does not exist.
        x += xj*qj;

        // Update z and gamma for the next iteration, except when j == c - 1.
        if (j < c - 1)
        {
            z = mul64s(z, powmod64s(gamma, xj, p), p);
            gamma = powmod64s(gamma, q, p);
            k /= q;
            qj *= q;
        }
    }
    return x;  // x := x_0 + x_1*q + x_2*q^2 + ... + x_{c - 1}*q^{c - 1};
}

/*
 * Sets up several variables and other initial overhead for log64s().
 *
 * PRECONDITION:
 *     Assumes that T has been allocated enough memory via computeSpace().
 *
 * NOTE:
 *     We assume that alpha and p (and thus n, Q, C, and k) are fixed. Hence, we
 *     want to precompute anything that depends only on these fixed values (that
 *     is, anything that is independent of y).
 *
 * INPUT:
 *     alpha: The base of the logarithm.
 *         p: The size of the finite field Z_p.
 *         n: The order of alpha in Z_p.
 *         Q: The primes in the prime factorization of n.
 *         C: The corresponding exponents of the primes in Q.
 *         k: The size of Q and C (the number of distinct primes of n).
 *         T: Temporary memory (gets written in this method once).
 */
void setupLog64s(LONG alpha, LONG p, LONG n, int* Q, int* C, int k, LONG* T)
{
    int i, j, q, c, s, t;
    LONG qc, alphaInv, m, beta, delta;
    
    int*  newK;
    LONG* newQ;
    LONG* newC;
    LONG*   QC;  // The array for each CRT modulus of the form q^c.
    LONG*    M;  // The array for each m = n/q^c needed for chrem64s().
    LONG*    I;  // The inverses needed for chrem64s().
    LONG*    G;  // The array for each gamma = beta^{-1} = alpha^{-m}.
    LONG*    D;  // The array for each delta = beta^{q^{c - 1}} = alpha^{n/q}.
    LONG**   Y;  // An array of k pointers to k hash tables for each alpha^{js}.
    int**    X;  // An array of k pointers to k hash tables for each js.
    LONG*  end;  // Points to the end (the first empty free slot in T).
    T[0]    = p;
    T[1]    = n;
    newK    = (int*)(T + 2);
    newK[0] = k;
    newQ    = (LONG*)(newK + 1);
    newC    = newQ + k;
    QC      = newC + k;
    M       = QC + k;
    I       = M + k;
    G       = I + k;
    D       = G + k;
    Y       = (LONG**)(D + k);
    X       = (int**)(Y + k);
    end     = (LONG*)(X + k);
    
    // Precompute the q^c moduli for chrem64s().
    for (i = 0; i < k; i++)
    {
        q = Q[i];
        c = C[i];
        qc = q;
        for (j = 1; j < c; j++)
            qc *= q;
        QC[i] = qc;
        newQ[i] = q;
        newC[i] = c;
    }
    
    // Using QC, precompute M and I for chrem64s().
    setupChrem64s(QC, n, M, I, k);
    
    // Using M, precompute each gamma and delta for pohligHellman64s().
    // Then by using Q and delta, precompute each hash table for logShanks64s().
    alphaInv = modinv64s(alpha, p);
    for (i = 0; i < k; i++)
    {
        // m     := n/q^c;
        // beta  := alpha^m = alpha^{n/q^c};
        // gamma := beta^{-1} = alpha^{-m};
        // delta := beta^{q^{c - 1}} = alpha^{n/q};
        q     = Q[i];
        m     = M[i];
        beta  = powmod64s(alpha, m, p);
        G[i]  = powmod64s(alphaInv, m, p);
        delta = powmod64s(beta, QC[i]/q, p);
        D[i]  = delta;
        if (q >= SHANKS_THRESHOLD)
        {
            s = ceilsqrt32s(q);
            t = 2*s + 1;
            Y[i] = end;
            X[i] = (int*)(Y[i] + t);
            end  = (LONG*)(X[i] + t);
            hashShanks64s(delta, p, q, Y[i], X[i]);
        }
    }
}

/*
 * Computes x = log[alpha](y) in Z_p via the Pohlig-Hellman algorithm.
 *
 * PRECONDITION:
 *     Assumes that setupLog64s() has already been called to set up T.
 *
 * INPUT:
 *     y: The argument of the logarithm (y = alpha^x).
 *     T: Temporary memory (read-only).
 *
 * INTENDED USAGE:
 *     Suppose that we are given alpha, p, n, Q, C, k, and an array Y of m
 *     logarithm arguments to be computed. Then we want to call the functions in
 *     the following order:
 *         int space = computeSpace(Q, k);
 *         LONG* T = (LONG*) arrayofbytes(space);
 *         setupLog64s(alpha, p, n, Q, C, k, T);
 *         int i;
 *         for (i = 0; i < m; i++)
 *             log64s(Y[i], T);
 *         free(T);
 */
LONG log64s(LONG y, LONG* T)
{
    int i;
    LONG x;
    LONG U[NUM_UNSIGNED_BITS];  // The array for each log[alpha](y) mod q^c.
    
    LONG   p;  // The size of the finite field Z_p.
    LONG   n;  // The order of alpha in Z_p.
    int*  kp;
    int    k;  // The number of distinct primes of n.
    LONG*  Q;  // The primes in the prime factorization of n.
    LONG*  C;  // The corresponding exponents of the primes in Q.
    LONG* QC;  // The array for each CRT modulus of the form q^c.
    LONG*  M;  // The array for each m = n/q^c needed for chrem64s().
    LONG*  I;  // The inverses needed for chrem64s().
    LONG*  G;  // The array for each gamma = beta^{-1} = alpha^{-m}.
    LONG*  D;  // The array for each delta = beta^{q^{c - 1}} = alpha^{n/q}.
    LONG** Y;  // An array of k pointers to k hash tables for each alpha^{js}.
    int**  X;  // An array of k pointers to k hash tables for each js.
    p   = T[0];
    n   = T[1];
    kp  = (int*)(T + 2);
    k   = kp[0];
    Q   = (LONG*)(kp + 1);
    C   = Q + k;
    QC  = C + k;
    M   = QC + k;
    I   = M + k;
    G   = I + k;
    D   = G + k;
    Y   = (LONG**)(D + k);
    X   = (int**)(Y + k);
    
    for (i = 0; i < k; i++)
    {
        //----------------------------------------------------------------------
        // Solving alpha^x = y   mod q^c in Z_p for 0 <= x < n is equivalent to
        // solving  beta^x = y^m mod q^c in Z_p for 0 <= x < q^c, where beta =
        // alpha^m and m = n/q^c. This reduces the number of multiplications in
        // each initial powmod64s() call in the for loop in pohligHellman64s()
        // from O(log n) to O(log q^c), since |alpha| = n and |beta| = q^c.
        //----------------------------------------------------------------------
        x = pohligHellman64s(powmod64s(y, M[i], p), D[i], p, G[i], QC[i], Q[i],
                             C[i], Y[i], X[i]);
        if (x == FAIL)
            return FAIL;
        U[i] = x;
    }
    return chrem64s(U, QC, n, M, I, k);
}

//------------------------------------------------------------------------------
// Memory management.
//------------------------------------------------------------------------------

/*
 * Computes the total space that must be allocated in bytes for setupLog64s().
 */
int computeSpace(int* Q, int k)
{
    int i, space, q, s, t;
    space  =      NUM_INT_VALS      * sizeof(int);
    space +=      NUM_LONG_VALS     * sizeof(LONG);
    space += k * (NUM_LONG_ARRS     * sizeof(LONG));
    space += k * (NUM_INT_PTR_ARRS  * sizeof(int*));
    space += k * (NUM_LONG_PTR_ARRS * sizeof(LONG*));
    for (i = 0; i < k; i++)
    {
        q = Q[i];
        if (q >= SHANKS_THRESHOLD)
        {
            s = ceilsqrt32s(q);
            t = 2*s + 1;
            space += t * sizeof(LONG);
            space += t * sizeof(int);
        }
    }
    return space;
}
